/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <tiuser.h>
#include <fcntl.h>
#include <signal.h>

extern int t_errno;
extern char *t_errlist[];

extern void log(const char *, ...);

#define tlog(STR) log("%s: %s", (STR), t_errlist[t_errno])

int
getfd(char *transport)
{
	int fd;

	/*
	 * Open the transport provider.
	 */
	if ((fd = t_open(transport, O_RDWR, NULL)) < 0)
		return(-1);

	/*
	 * Bind an arbitrary address to the transport
	 * endpoint.
	 */
	if (t_bind(fd, NULL, NULL) < 0) {
		t_close(fd);
		return(-1);
	}
	return(fd);
}

void
serve(int fd, char *transport)
{
	int afd;
	pid_t pid;
	struct t_call *call;

	/*
	 * Allocate a t_call structure to be used when
	 * receiving connect indications.
	 */
	call = (struct t_call *)t_alloc(fd, T_CALL, T_ALL);
	if (call == NULL) {
		tlog("t_alloc failed");
		exit(1);
	}
	for ( ; ; ) {
		/*
		 * Block until a connect indication arrives.
		 */
		if (t_listen(fd, call) < 0) {
			tlog("t_listen failed");
			exit(1);
		}

		/*
		 * Get a new transport endpoint to be used in
		 * the connection.
		 */
		if ((afd = getfd(transport)) < 0) {
			/*
			 * Can't get a new endpoint -- reject
			 * the connect request.
			 */
			if (t_snddis(fd, call) < 0) {
				tlog("t_snddis failed");
				exit(1);
			}
			continue;
		}

		/*
		 * Accept the connection on afd instead of fd.
		 */
		if (t_accept(fd, afd, call) < 0) {
			if (t_errno == TLOOK) {
				/*
				 * Someone canceled the connect request.
				 */
				if (t_rcvdis(fd, NULL) < 0) {
					tlog("t_rcvdis failed");
					exit(1);
				}
			} else {
				tlog("t_accept failed");
				exit(1);
			}
		} else {
			/*
			 * Fork and let the child process do the real
			 * work.  The parent goes back to listen for
			 * more connect indications.
			 */
			if ((pid = fork()) < 0) {
				/*
				 * Fork failed -- reject
				 * the connect request.
				 */
				log("fork failed");
				if (t_snddis(afd, NULL) < 0) {
					tlog("t_snddis failed");
					exit(1);
				}
			} else if (pid == 0) {
				/*
				 * Child resumes here.  Don't need
				 * the file descriptor used to listen
				 * for connect indications.
				 */
				t_close(fd);

				/*
				 * Perform the service using afd and exit.
				 */
			}
		}

		/*
		 * After a successful fork, the parent continues
		 * here. Close the file descriptor used in the
		 * connection and continue to listen for more
		 * connect indications.
		 */
		t_close(afd);
	}
}
