/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
/* netty.c */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mkdev.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#define DEVMODE	(S_IRUSR|S_IWUSR)
#define FATAL(s)	{ log((s), strerror(errno)); exit(1); }

extern int initlog(const char *);
extern void log(const char *, ...);

void
main()
{
	char *name;
	char devname[64];
	struct stat sbuf;
	dev_t dev;
	int fd;

	/*
	 * Open the log file.
	 */
	if (initlog("/var/tmp/netty.log") < 0)
		exit(1);

	/*
	 * Find the device number of the standard input,
	 * which is connected to the network stream.
	 */
	if (fstat(0, &sbuf) < 0)
		FATAL("cannot stat stdin");

	/*
	 * Get the transport provider's device name
	 * prefix.
	 */
	name = getenv("MPREFIX");
	if (name == NULL) {
		log("cannot find provider's device name");
		exit(1);
	}

	/*
	 * Create the device filename corresponding to the
	 * network connection.
	 */
	dev = sbuf.st_rdev;
	sprintf(devname, "/dev/%s%03d", name, minor(dev));

	/*
	 * If the device node is not there, try to create
	 * it.  Otherwise, change the owner to the super-
	 * user and make it accessible only by the owner.
	 * login will change this later.
	 */
	if (stat(devname, &sbuf) < 0) {
		if (mknod(devname, S_IFCHR|DEVMODE, dev) < 0)
			FATAL("cannot make node: %s");
	} else {
		if (chown(devname, 0, 0) < 0)
			FATAL("cannot change device ownership: %s");
		if (chmod(devname, DEVMODE) < 0)
			FATAL("cannot change device mode: %s");
	}

	/*
	 * Now open the device node.
	 */
	if ((fd = open(devname, O_RDWR)) < 0)
		FATAL("cannot open device: %s");

	/*
	 * Close file descriptors 0, 1, and 2.  Since
	 * we've just reopened the network device, the
	 * connection will remain intact.
	 */
	close(0);
	close(1);
	close(2);

	/*
	 * Now make file descriptors 0, 1, and 2 into
	 * copies of the file descriptor of the device node.
	 * Then close the original fd open to the node.
	 */
	if ((dup(fd) != 0) || (dup(fd) != 1) ||
	  (dup(fd) != 2))
		FATAL("cannot set up file descriptors: %s");
	close(fd);

	/*
	 * Finally exec login to complete the remote
	 * login process.
	 */
	execl("/usr/bin/login", "login", "-d", devname, NULL);
	FATAL("cannot exec login: %s");
}
